/*
 * arch/arm/mach-ep93xx/board-trumpeter.c
 *
 * Support for Bluewater Systems Trumpeter baseboard
 * 
 * Author:	Ryan Mallon <ryan@bluewatersys.com>
 * Copyright:	2009 Bluewater Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/serial_reg.h>
#include <linux/serial_8250.h>
#include <linux/irq.h>
#include <linux/i2c.h>
#include <linux/i2c/pca953x.h>

#include <asm/io.h>
#include <asm/hardware/snapper.h>

#include <mach/platform.h>
#include <mach/ep93xx-regs.h>
#include <mach/gpio.h>
#include <mach/irqs.h>

#define QUART_PORT(_base, _irq)						\
	{								\
		.irq		= gpio_to_irq(_irq),			\
		.uartclk	= 14745600,				\
		.iotype		= UPIO_MEM,				\
		.flags		= UPF_BOOT_AUTOCONF | UPF_IOREMAP,	\
		.regshift	= 2,					\
		.mapbase	= _base,				\
	}

static struct plat_serial8250_port quad_uart_data[] = {
	/* Quad UART A */
	QUART_PORT(EP93XX_CS2_PHYS_BASE + 0x00, EP93XX_GPIO_LINE_A(6)),
	QUART_PORT(EP93XX_CS2_PHYS_BASE + 0x40, EP93XX_GPIO_LINE_A(6)),
	QUART_PORT(EP93XX_CS2_PHYS_BASE + 0x80, EP93XX_GPIO_LINE_A(6)),
	QUART_PORT(EP93XX_CS2_PHYS_BASE + 0xc0, EP93XX_GPIO_LINE_A(6)),

	/* Quad UART B */
	QUART_PORT(EP93XX_CS3_PHYS_BASE + 0x00, EP93XX_GPIO_LINE_B(1)),
	QUART_PORT(EP93XX_CS3_PHYS_BASE + 0x40, EP93XX_GPIO_LINE_B(1)),
	QUART_PORT(EP93XX_CS3_PHYS_BASE + 0x80, EP93XX_GPIO_LINE_B(1)),
	QUART_PORT(EP93XX_CS3_PHYS_BASE + 0xc0, EP93XX_GPIO_LINE_B(1)),
	{ },
};

static struct platform_device quad_uart = {
	.name	= "serial8250",
	.id	= PLAT8250_DEV_PLATFORM,
	.dev	= {
		.platform_data = quad_uart_data,
	},
};

static const char *trumpeter_max7311_names[2][16] = {
	{"tm_485", "hb1_485", "hb2_485", "pc1_485", "pc2_485", "tm_shdn_n",
	 "hb1_shdn_n", "hb2_shdn_n", "pc1_shdn_n", "pc2_shdn_n", 
	 "sys_rst_gen_n", "3g_disable_n", "fg0_scl_en_n", "fg1_scl_en_n",
	 "pwr_halt_n", "pwr_button_n"},
	
	{"mdt_ready", "mdt_invalid_n", "mdt_forceoff_n", "mdt_forceon",
	 "pmr_ready", "pmr_invalid_n", "pmr_forceoff_n", "pmr_forceon",
	 "gps_ready", "gps_invalid_n", "gps_forceoff_n", "gps_forceon",
	 "acp_shdn_n", "wlan_pd_n", "gsm_reset_n", "psm_pwr_en_n",
	}
};

static int trumpeter_max7311_setup(struct i2c_client *client, unsigned base,
				   unsigned ngpio, void *context)
{
	unsigned long config = (unsigned long)context;
	int i, exp, gpio, val, err;
	
	for (i = 0; i < ngpio; i++) {
		gpio = base + i;
		val = gpio_get_value(gpio);
		
		exp = base == NR_BUILTIN_GPIO ? 0 : 1;
		err = gpio_request(gpio, 
				   trumpeter_max7311_names[exp][gpio - base]);
		if (err)
			goto fail;

		if (config & (1 << i))
			gpio_direction_input(gpio);
		else
			gpio_direction_output(gpio, val);
	}

	return 0;

fail:
	while (--i >= 0)
		gpio_free(base + i);
	return err;
}

static int trumpeter_max7311_teardown(struct i2c_client *client, unsigned base,
				      unsigned ngpio, void *context)
{
	int i;

	for (i = 0; i < ngpio; i++)
		gpio_free(base + i);

	return 0;
}

static struct pca953x_platform_data trumpeter_max7311_data[] = {
	[0] = {
		.gpio_base	= NR_BUILTIN_GPIO,
		.context	= (void *)0x8400,
		.setup		= trumpeter_max7311_setup,
		.teardown	= trumpeter_max7311_teardown,	
	},
	[1] = {
		.gpio_base	= NR_BUILTIN_GPIO + 16,
		.context	= (void *)0x0333,
		.setup		= trumpeter_max7311_setup,
		.teardown	= trumpeter_max7311_teardown,	
	},
};

static struct i2c_board_info __initdata trumpeter_i2c_data[] = {
	{
		I2C_BOARD_INFO("max7311", 0x20),
		.platform_data	= &trumpeter_max7311_data[0],
	},
	{
		I2C_BOARD_INFO("max7311", 0x27),
		.platform_data	= &trumpeter_max7311_data[1],
	},
	{
		I2C_BOARD_INFO("ds2782", 0x35),
	},
	{
		I2C_BOARD_INFO("ds2782", 0x36),
	},
};

static int __init trumpeter_init(void)
{
	/* I2C Devices */
	ep93xx_register_i2c(trumpeter_i2c_data, 
			    ARRAY_SIZE(trumpeter_i2c_data));
	
	/* Timings for Quad UART chip selects (horrendously slow) */
	__raw_writel(0x3000ffef, EP93XX_SMCBCR(2));
	__raw_writel(0x3000ffef, EP93XX_SMCBCR(3));
	set_irq_type(gpio_to_irq(EP93XX_GPIO_LINE_A(6)), IRQF_TRIGGER_LOW);
	set_irq_type(gpio_to_irq(EP93XX_GPIO_LINE_B(1)), IRQF_TRIGGER_LOW);
	gpio_request(EP93XX_GPIO_LINE_A(6), "quad_uart_a");
	gpio_request(EP93XX_GPIO_LINE_B(1), "quad_uart_b");
	platform_device_register(&quad_uart);

	return 0;
}

SNAPPER_BASEBOARD(trumpeter, trumpeter_init);
