/*
 * Simple AT91SAM9260 touchscreen driver
 *
 * Author: Andrew Turner <aturner@bluewatersys.com>
 *
 * Copyright (c) 2009, Bluewater Systems Ltd
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/init.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/workqueue.h>
#include <linux/platform_device.h>

#include <mach/gpio.h>

#define	MAX_10BIT	((1 << 10) - 1)

#define TOUCH_YP	AT91_PIN_PC2
#define TOUCH_XP	AT91_PIN_PC3
#define TOUCH_XN	AT91_PIN_PA26
#define TOUCH_YN	AT91_PIN_PA27

struct sn9260_ts {
	struct at91_adc		*adc;
	struct input_dev	*input;
	struct timer_list	timer;
	struct work_struct	work;
	char phys[32];
};

extern int at91_adc_get_value(struct at91_adc *adc, int channel);

static void sn9260_ts_work(struct work_struct *work)
{
	struct sn9260_ts *ts = container_of(work, struct sn9260_ts, work);
	int touch, x, y;

	/* Is the user touching the screen */
	at91_set_A_periph(TOUCH_YP, 0);
	at91_set_gpio_output(TOUCH_YN, 0);
	at91_set_gpio_output(TOUCH_XP, 1);
	at91_set_gpio_output(TOUCH_XN, 0);

	touch = at91_adc_get_value(ts->adc, 2);
	input_report_abs(ts->input, ABS_PRESSURE, touch > 20 ? touch : 0);
	input_report_key(ts->input, BTN_TOUCH, touch > 20);

	if (touch > 20) {
		/* Read the X position */
		at91_set_A_periph(TOUCH_YP, 0);
		at91_set_gpio_input(TOUCH_YN, 0);
		at91_set_gpio_output(TOUCH_XP, 1);
		at91_set_gpio_output(TOUCH_XN, 0);
		x = at91_adc_get_value(ts->adc, 2);

		/* Read the Y position */
		at91_set_gpio_output(TOUCH_YP, 1);
		at91_set_gpio_output(TOUCH_YN, 0);
		at91_set_A_periph(TOUCH_XP, 0);
		at91_set_gpio_input(TOUCH_XN, 0);
		y = at91_adc_get_value(ts->adc, 3);

		input_report_abs(ts->input, ABS_X, x);
		input_report_abs(ts->input, ABS_Y, y);
	}

	input_sync(ts->input);
}

static void sn9260_ts_timer(unsigned long data)
{
	struct sn9260_ts *ts = (struct sn9260_ts *)data;

	schedule_work(&ts->work);
	mod_timer(&ts->timer, jiffies + HZ / 100);
}

/* TODO: Move to the at91_adc driver */
static int at91_adc_find(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	const char *name = data;

	return (strcmp(pdev->name, name) == 0);
}

static struct at91_adc *at91_adc_get(void)
{
	struct device *dev;
	
	dev = bus_find_device(&platform_bus_type, NULL, 
			      "at91_adc", at91_adc_find);
	if (!dev)
		return NULL;
	
	return dev_get_drvdata(dev);
}

static int __devinit sn9260_ts_probe(struct platform_device *pdev)
{
	struct sn9260_ts	*ts;
	struct input_dev	*input_dev;
	int			err;

	ts = kzalloc(sizeof(struct sn9260_ts), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (ts == NULL || input_dev == NULL) {
		err = -ENOMEM;
		goto err_free_mem;
	}
	platform_set_drvdata(pdev, ts);

	ts->adc = at91_adc_get();
	if (ts->adc == NULL) {
		err = -ENXIO;
		goto err_free_mem;
	}

	snprintf(ts->phys, sizeof(ts->phys),
		 "%s/input0", dev_name(&pdev->dev));

	ts->input = input_dev;
	input_dev->name = "Snapper9260 Touchscreen";
	input_dev->phys = ts->phys;
	input_dev->id.bustype = BUS_HOST;
	input_dev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
	input_dev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);

	input_set_abs_params(input_dev, ABS_X, 0, MAX_10BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_Y, 0, MAX_10BIT, 0, 0);
	input_set_abs_params(input_dev, ABS_PRESSURE, 0, MAX_10BIT, 0, 0);

	err = input_register_device(ts->input);
	if (err)
		goto err_free_mem;

	init_timer(&ts->timer);
	ts->timer.data = ts;
	ts->timer.function = sn9260_ts_timer;

	INIT_WORK(&ts->work, sn9260_ts_work);
	mod_timer(&ts->timer, jiffies + HZ / 100);

	return 0;

err_free_mem:
	input_free_device(input_dev);
	kfree(ts);
	return err;
}

static int __devexit sn9260_ts_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver sn9260_ts_driver = {
	.probe		= sn9260_ts_probe,
	.remove		= __devexit_p(sn9260_ts_remove),
	.driver		= {
		.name	= "sn9260_ts",
	},
};

static int __init sn9260_ts_init(void)
{
	return platform_driver_register(&sn9260_ts_driver);
}

static void __exit sn9260_ts_exit(void)
{
	platform_driver_unregister(&sn9260_ts_driver);
}

module_init(sn9260_ts_init);
module_exit(sn9260_ts_exit);

MODULE_AUTHOR("Andrew Turner <aturner@bluewatersys.com>");
MODULE_DESCRIPTION("AT91SAM9260 TouchScreen Driver");
MODULE_LICENSE("GPL");

