/*
 *  drivers/mtd/nand/snappercl15_nand.c
 *
 *  Copyright (C) 2009 Andre Renaud (andre@bluewatersys.com)
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   Snapper CL15 modules. The specific chip varies, and some old modules
 *   used xD card in this socket (which has the same NAND interface).
 */

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>

//#define USE_DEBUG

#ifdef USE_DEBUG
#define DBG(a,b...) printk ("%s: %d: " a "\n", __FUNCTION__, __LINE__ , ##b)
#else
#define DBG(a,b...)
#endif

#define NAME "CL15 NAND"

enum {
        NAND_wpn = 1 << 8,
        NAND_ale = 1 << 9,
        NAND_cle = 1 << 10,
        NAND_cen = 1 << 11,

        NAND_rdy = 1 << 14,
        NAND_gndcd = 1 << 15,
};

static struct
{
	unsigned int state;
	volatile u16 *data;
	volatile u16 *ctrl;
} cl15_nand;

static struct mtd_info *cl15_mtd = NULL;

static inline void nand_bic_or (unsigned int bic, unsigned int or)
{
	cl15_nand.state &= ~bic;
	cl15_nand.state |= or;
	*cl15_nand.ctrl = cl15_nand.state;
	DBG ("bic_or 0x%x 0x%x high: %x", bic, or, cl15_nand.state);
}

/*
 * Define partitions for flash device
 */
const static struct mtd_partition partition_info[] = {
   { name: "Kernel",
     offset: 0,
     size: 2 * 1024 * 1024 },
   { name: "Filesystem",
     offset: 2 * 1024 * 1024,
     size: MTDPART_SIZ_FULL },
   { name: "Whole",
     offset: 0,
     size: MTDPART_SIZ_FULL}
};

#define NUM_PARTITIONS (sizeof (partition_info) / sizeof (partition_info[0]))

static inline void cl15_write_byte(struct mtd_info *mtd, u_char byte)
{
	int i = byte & 0xff;

	DBG ("i=%x", i);
	*cl15_nand.data = cl15_nand.state | i;
}

static void cl15_hwcontrol (struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	if (ctrl & NAND_CTRL_CHANGE) {
		unsigned int set = NAND_cen;
		if (ctrl & NAND_NCE)
			set &= ~NAND_cen;
		if (ctrl & NAND_CLE)
			set |= NAND_cle;
		if (ctrl & NAND_ALE)
			set |= NAND_ale;
		nand_bic_or (NAND_cen | NAND_cle | NAND_ale, set);
	}

	if (cmd != NAND_CMD_NONE)
		cl15_write_byte(mtd, cmd);
}

/* Dummy function. Necessary to prevent kernel panics */
static void cl15_select_chip(struct mtd_info *mtd, int chip)
{
}

static u_char cl15_read_byte(struct mtd_info *mtd)
{
	int data;
	data = *cl15_nand.data & 0xff;
	DBG ("data=%x", data);
	return (u_char) data;
}

/* 
 * We can't used writesb, as NAND_wpn is always set, which we'd be clearing
 * in that case
 */
static void cl15_write_buf(struct mtd_info* mtd, const u_char *buf, int len)
{
        while (len--)
                *cl15_nand.data = cl15_nand.state | *buf++;
}

static void cl15_read_buf(struct mtd_info* mtd, u_char *buf, int len)
{
        readsb((uint8_t *)cl15_nand.data, buf, len);
}

static int cl15_verify_buf(struct mtd_info* mtd, const u_char* buf, int len)
{
	int i;
	u_char data;
	for (i=0; i<len; i++) {
                data = *cl15_nand.data & 0xff;
		if (buf[i] != data) {
			DBG ("Verify %d diff: %x != %x\n", i, buf[i], data);
			return -EFAULT;
		}
	}

	return 0;
}

static int cl15_dev_ready(struct mtd_info *mtd)
{
	return (*cl15_nand.ctrl & NAND_rdy) != 0;
}

static int cl15_card_scan (void)
{
	struct nand_chip *this;

	/* Allocate memory for MTD device structure and private data */
	cl15_mtd = kmalloc (sizeof(struct mtd_info) + sizeof (struct nand_chip),
			GFP_KERNEL);
	if (!cl15_mtd) {
		printk (KERN_ERR NAME ": Unable to allocate Snapper Cl15 NAND MTD device structure.\n");
		return -ENOMEM;
	}
	/* Get pointer to private data */
	this = (struct nand_chip *) (&cl15_mtd[1]);

	/* Initialize structures */
	memset((char *) cl15_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));
	/* Link the private data with the MTD structure */
	cl15_mtd->priv = this;
	cl15_mtd->name = NAME;
	cl15_mtd->owner = THIS_MODULE;
	/* Register nand functions*/
	this->cmd_ctrl = cl15_hwcontrol;
	this->select_chip = cl15_select_chip;	
	this->chip_delay = 25;
	this->read_byte = cl15_read_byte;
	this->dev_ready = cl15_dev_ready;
	this->read_buf = cl15_read_buf;
	this->write_buf = cl15_write_buf;
	this->verify_buf = cl15_verify_buf;

	this->options = NAND_NO_AUTOINCR;

	this->ecc.mode = NAND_ECC_SOFT;

	/* Setup nand addresses */
	this->IO_ADDR_R = 0;
	this->IO_ADDR_W = 0;
	cl15_nand.state =0;

        /* FIXME: Are the NAND timings in EP93XX_SMCBCR(7) as fast as possible? */
        nand_bic_or (0, NAND_cen | NAND_wpn);

        /* Scan to find existence of the device */
	if (nand_scan (cl15_mtd, 1)) {
		kfree (cl15_mtd);
		cl15_mtd = NULL;
		return -ENODEV;
	}

	/* Register the partitions */
	add_mtd_partitions(cl15_mtd, partition_info, NUM_PARTITIONS);

        return 0;
}

static int __init cl15_init (void)
{
        int r;

	printk(NAME ": init\n");

	cl15_nand.state = 0;
	cl15_nand.data = (volatile u16 *) ioremap(0x71000000,0x1000);
	if (!cl15_nand.data) {
	   printk (KERN_ERR NAME ": Unable to remap nand data\n");
	   return -ENOMEM;
	}
	cl15_nand.ctrl = (volatile u16 *)((int)cl15_nand.data + 0x40);

        r = cl15_card_scan ();
        if (r < 0) {
                iounmap(cl15_nand.data);
                cl15_nand.ctrl = cl15_nand.data = NULL;
                return r;
        }
	
	return 0;
}
module_init(cl15_init);
/*
 * Clean up routine
 */
#ifdef MODULE
static void __exit cl15_cleanup (void)
{
    	if (cl15_mtd)
	{
		/* Release resources, unregister device */
		nand_release (cl15_mtd);

		/* Free the MTD device structure */
		kfree (cl15_mtd);
	}

	iounmap((void *)cl15_nand.data);
}
module_exit(cl15_cleanup);
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("James R. McKaskill <james@bluewatersys.com>");
MODULE_DESCRIPTION("Device layer for NAND on Snapper CL15");

