/*
 * drivers/pcmcia/ep93xx_salmon.h
 *
 * PCMCIA support for the Snapper CL15 Salmon baseboard
 * 
 * Author:	Ryan Mallon <ryan@bluewatersys.com>
 * Copyright:	2009 Bluewater Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/jiffies.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>

#include <asm/hardware/snapper.h>
#include <mach/hardware.h>
#include <mach/salmon.h>
#include <mach/irqs.h>
#include <asm/irq.h>

#include "ep93xx_pcmcia.h"

#define SALMON_CF_DETECT		EP93XX_GPIO_LINE_D(7)
#define SALMON_S1_GRAB_WIFI		EP93XX_GPIO_LINE_D(6)
#define SALMON_WIFI_ENABLE		EP93XX_GPIO_LINE_G(2)

static int socket_irq[] = {gpio_to_irq(EP93XX_GPIO_LINE_F(6))};

static void salmon_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				       struct pcmcia_state *state)
{
	unsigned long status;

	status = __raw_readl(EP93XX_GPIO_PFDR);	
	state->detect = !gpio_get_value(SALMON_CF_DETECT);
	
	/* Module S1 needs Wifi grab enabled in addition to card detect */
	if (salmon_is_module_s1() && !gpio_get_value(SALMON_S1_GRAB_WIFI))
		state->detect = 0;

	state->ready	= (status & EP93XX_PCMCIA_READY)  ? 1 : 0;
	state->wrprot	= (status & EP93XX_PCMCIA_WRPROT) ? 0 : 1;
	state->vs_3v 	= 1;
	state->bvd1	= 0;
	state->bvd2	= 0;
}

static int salmon_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
					  const socket_state_t *state)
{
	gpio_set_value(SALMON_WIFI_ENABLE, state->Vcc);
	return ep93xx_pcmcia_configure_socket(skt, state);
}

static int salmon_pcmcia_probe(struct device *dev)
{
	int err;

	err = gpio_request(SALMON_CF_DETECT, "CF Detect");
	if (err)
		goto fail_gpio_cf_detect;
	
	err = gpio_request(SALMON_S1_GRAB_WIFI, "Grab Wifi");
	if (err)
		goto fail_gpio_grab_wifi;

	err = gpio_request(SALMON_WIFI_ENABLE, "Wifi Enable");
	if (err)
		goto fail_gpio_wifi_enable;

	return 0;

fail_gpio_wifi_enable:
	gpio_free(SALMON_S1_GRAB_WIFI);
fail_gpio_grab_wifi:
	gpio_free(SALMON_CF_DETECT);
fail_gpio_cf_detect:
	return err;		
}

static struct pcmcia_low_level salmon_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= ep93xx_pcmcia_hw_init,
	.hw_shutdown		= ep93xx_pcmcia_hw_shutdown,
	.socket_state		= salmon_pcmcia_socket_state,
	.configure_socket	= salmon_pcmcia_configure_socket,
	.socket_init		= ep93xx_pcmcia_socket_init,
	.socket_suspend		= ep93xx_pcmcia_socket_suspend,
	.first			= 0,
	.nr			= 1,
};

static int __init salmon_pcmcia_init(void)
{
	if (!snapper_baseboard_is_salmon()) 
		return -ENODEV;
	
	if (gpio_get_value(SALMON_CF_DETECT)) {
		printk(KERN_ERR "salmon: Wifi module not present!\n");
		return -ENODEV;
	}

	ep93xx_pcmcia_set_irqs(socket_irq, NULL, 0);
	return ep93xx_pcmcia_probe(salmon_pcmcia_probe, &salmon_pcmcia_ops);
}

static void __exit salmon_pcmcia_exit(void)
{
	ep93xx_pcmcia_unregister();
}

module_init(salmon_pcmcia_init);
module_exit(salmon_pcmcia_exit);

MODULE_AUTHOR("Ryan Mallon <ryan@bluewatersys.com>");
MODULE_DESCRIPTION("Snapper CL15/Salmon compact flash driver");
MODULE_LICENSE("GPL");

