/*
 * Cirrus EP93xx SPI Driver
 *
 * Copyright (c) 2007 Ryan Mallon <ryan@bluewatersys.com>
 * Copyright (c) 2009 H Hartley Sweeten <hsweeten@visionengravers.com>
 *
 * Based on the ep93xx spi bitbang driver by Chase Douglas
 * and linux/drivers/spi/spi_txx9.c
 *
 * Licensed under the GPL-2 or later.
 */

#include <linux/platform_device.h>
#include <linux/workqueue.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/spi/spi.h>

#include <mach/spi.h>

/*
 * Synchronous Serial Port Register offsets
 */
#define SSPCR0			0x00
#define SSPCR1			0x04
#define SSPDR			0x08
#define SSPSR			0x0c
#define SSPCPSR			0x10
#define SSPIIR			0x14
#define SSPICR			SSPIIR

/* Control Register 0 bit defines */
#define SSPCR0_SCR_MASK		(0x0000ff00)
#define SSPCR0_SCR_SHIFT	(8)
#define SSPCR0_SPH		(1<<7)
#define SSPCR0_SPO		(1<<6)
#define SSPCR0_FRF_MASK		(0x00000030)
#define SSPCR0_FRF_SHIFT	(4)
#define SSPCR0_FRF_MOTOROLA	(0<<4)
#define SSPCR0_FRF_TI		(1<<4)
#define SSPCR0_FRF_NI		(2<<4)
#define SSPCR0_DSS_MASK		(0x0000000f)
#define SSPCR0_DSS_SHIFT	(0)
#define SSPCR0_DSS_4BIT		(3<<0)
#define SSPCR0_DSS_5BIT		(4<<0)
#define SSPCR0_DSS_6BIT		(5<<0)
#define SSPCR0_DSS_7BIT		(6<<0)
#define SSPCR0_DSS_8BIT		(7<<0)
#define SSPCR0_DSS_9BIT		(8<<0)
#define SSPCR0_DSS_10BIT	(9<<0)
#define SSPCR0_DSS_11BIT	(10<<0)
#define SSPCR0_DSS_12BIT	(11<<0)
#define SSPCR0_DSS_13BIT	(12<<0)
#define SSPCR0_DSS_14BIT	(13<<0)
#define SSPCR0_DSS_15BIT	(14<<0)
#define SSPCR0_DSS_16BIT	(15<<0)

/* Control Register 1 bit defines */
#define SSPCR1_SOD		(1<<6)
#define SSPCR1_MS		(1<<5)
#define SSPCR1_SSE		(1<<4)
#define SSPCR1_LBM		(1<<3)
#define SSPCR1_RORIE		(1<<2)
#define SSPCR1_TIE		(1<<1)
#define SSPCR1_RIE		(1<<0)

/* Receive (read) / Transmit (write) FIFO Register bit defines */
#define SSPDR_MASK		(0x0000ffff)
#define SSPDR_SHIFT		(0)

#define EP93XX_SSP_FIFO_SIZE	8

/* Status Register bit defines */
#define SSPSR_BSY		(1<<4)
#define SSPSR_RFF		(1<<3)
#define SSPSR_RNE		(1<<2)
#define SSPSR_TNF		(1<<1)
#define SSPSR_TFE		(1<<0)

/* Clock Prescale Register bit defines */
#define SSPCPSR_MASK		(0x000000ff)
#define SSPCPSR_SHIFT		(0)

/* Interrupt Identification/Clear (rd/wr) register bit defines */
#define SSPIIR_RORIS		(1<<2)
#define SSPIIR_TIS		(1<<1)
#define SSPIIR_RIS		(1<<0)

/*
 * Unique chip data for each slave device on the SPI bus (external drivers)
 */
struct ep93xx_chip_data {
	u32	max_speed_hz;
	u8	bits_per_word;
	int	cs_active_high;
	u8	mode;
	void	(*cs_control)(int active, int activehigh);
	void	(*cleanup)(void);
};

/*
 * Private data for the EP93xx SPI controller (this driver)
 */
struct ep93xx_spi {
	struct spi_master		*master;
	struct device			*dev;
	struct ep93xx_spi_master	*pdata;
	struct clk			*clk;
	void __iomem			*mmio_base;

	struct ep93xx_chip_data		*chip;

	struct list_head		msg_queue;
	struct workqueue_struct 	*workq;
	struct work_struct		work;
	wait_queue_head_t		waitq;

	spinlock_t			lock;
};


static inline unsigned int ep93xx_spi_readl(struct ep93xx_spi *info,
					    unsigned int off)
{
	return __raw_readl(info->mmio_base + off);
}

static inline void ep93xx_spi_writel(struct ep93xx_spi *info,
				     unsigned int val, unsigned int off)
{
	__raw_writel(val, info->mmio_base + off);
}

/*
 * Read a value from the Receive FIFO
 */
static inline unsigned int ep93xx_spi_read_data(struct ep93xx_spi *info)
{
	return ep93xx_spi_readl(info, SSPDR) & SSPDR_MASK;
}

/*
 * Write a value to the Transmit FIFO
 */
static inline void ep93xx_spi_write_data(struct ep93xx_spi *info,
					 unsigned int val)
{
	ep93xx_spi_writel(info, val & SSPDR_MASK, SSPDR);
}

/*
 * Check if the Transmit FIFO is full
 */
static inline int ep93xx_spi_tx_full(struct ep93xx_spi *info)
{
	return (ep93xx_spi_readl(info, SSPSR) & SSPSR_TNF) ? 0 : 1;
}

/*
 * Check it the Transmit FIFO is empty
 */
static inline int ep93xx_spi_tx_empty(struct ep93xx_spi *info)
{
	return (ep93xx_spi_readl(info, SSPSR) & SSPSR_TFE) ? 1 : 0;
}

/*
 * Check if the Receive FIFO is empty
 */
static inline int ep93xx_spi_rx_empty(struct ep93xx_spi *info)
{
	return (ep93xx_spi_readl(info, SSPSR) & SSPSR_RNE) ? 0 : 1;
}

/*
 * Enable the SSP controller
 */
static inline void ep93xx_spi_enable(struct ep93xx_spi *info)
{
	ep93xx_spi_writel(info, SSPCR1_SSE, SSPCR1);
}

/*
 * Disable the SSP controller
 */
static inline void ep93xx_spi_disable(struct ep93xx_spi *info)
{
	ep93xx_spi_writel(info, 0x0, SSPCR1);
}

/*
 * Get the next transfer from the list, lock is already held
 */
static inline struct spi_transfer * ep93xx_spi_next(struct spi_transfer *trans,
						    struct spi_message *msg)
{
	if (trans->transfer_list.next == &msg->transfers)
		return NULL;

	return list_entry(trans->transfer_list.next,
			  struct spi_transfer, transfer_list);
}

static void ep93xx_apply_spi_setup(struct ep93xx_spi *info, 
				   struct ep93xx_chip_data *chip, u32 speed_hz)
{
	unsigned int val;
	
	/*
	 * REVISIT: It appears the EP93xx SSP port must be enabled then
	 * disabled in order to update its configuration.
	 */
	ep93xx_spi_enable(info);

	/* Set the clock rate for the transfer */
	clk_set_rate(info->clk, speed_hz ? speed_hz : chip->max_speed_hz);

	val = ep93xx_spi_readl(info, SSPCR0);
	val &= SSPCR0_SCR_MASK;
	val |= SSPCR0_FRF_MOTOROLA;
#if 1
	/*
	 * FIXME: The following code should work but caused problems with
	 * the mmc_spi driver.
	 * Edit: this works ok, so long as MMC uses SPI MODE_3 instead of MODE_0
	 */
	val |= ((chip->mode & SPI_CPHA) ? SSPCR0_SPH : 0);
	val |= ((chip->mode & SPI_CPOL) ? SSPCR0_SPO : 0);
#else
	 /*
	 * For right now the following puts the controller in SPI_MODE_3.
	 */
	val |= SSPCR0_SPH | SSPCR0_SPO;
#endif
	val |= (chip->bits_per_word - 1);
	ep93xx_spi_writel(info, val, SSPCR0);
	
	ep93xx_spi_disable(info);

	/* Now enable the port for operation */
	ep93xx_spi_enable(info);
}

/*
 * Process all the transfers in a spi message
 */
static void ep93xx_spi_handle_message(struct ep93xx_spi *info,
				      struct spi_message *msg)
{
	struct spi_transfer *rx, *tx;
	int rx_left, tx_left, sent = 0, recv = 0;
	const u8 *tx_buf;
	u8 *rx_buf;
	unsigned long flags;

	spin_lock_irqsave(&info->lock, flags);
	rx = tx = list_entry(msg->transfers.next,
			struct spi_transfer, transfer_list);
	spin_unlock_irqrestore(&info->lock, flags);

	tx_left = tx->len;
	rx_left = rx->len;
	tx_buf = tx->tx_buf;
	rx_buf = rx->rx_buf;

	msg->actual_length = 0;

	/* FIXME: Do we need to change bits per word for the transfer? */
	if (tx->bits_per_word &&
	    tx->bits_per_word != info->chip->bits_per_word) {
		dev_dbg(info->dev,
			"transfer->bits_per_word = %d\n", tx->bits_per_word);
	}
	
	ep93xx_apply_spi_setup(info, info->chip, tx->speed_hz);

	/* Enable the frame latch */
	if (info->chip->cs_control)
		info->chip->cs_control(1, info->chip->cs_active_high);

	while (tx || rx) {

		/* Send */
		while (tx && !ep93xx_spi_tx_full(info)) {
			/* Don't send more bytes than we have read */
			if (sent - recv >= EP93XX_SSP_FIFO_SIZE)
				break;

			if (tx_left) {
				if (tx_buf) {
					ep93xx_spi_write_data(info, *tx_buf);
					tx_buf++;
				} else {
					ep93xx_spi_write_data(info, 0);
				}
				tx_left--;
				sent++;
			}

			if (!tx_left) {
				/* Next transfer */
				spin_lock_irqsave(&info->lock, flags);
				tx = ep93xx_spi_next(tx, msg);
				spin_unlock_irqrestore(&info->lock, flags);

				if (tx) {
					tx_left = tx->len;
					tx_buf = tx->tx_buf;
					if (tx->speed_hz) {
						/* FIXME: Change speed? */
						dev_dbg(info->dev,
							"transfer->speed_hz = %u\n",
							tx->speed_hz);
					}
				}
			}
		}

		/* Receive */
		while (rx && !ep93xx_spi_rx_empty(info)) {
			/* Don't empty the rx fifo too early */
			if (tx && (sent - recv <= EP93XX_SSP_FIFO_SIZE / 2))
				break;

			if (rx_left) {
				u8 data;

				data = ep93xx_spi_read_data(info);
				if (rx_buf) {
					*rx_buf = data;
					rx_buf++;
				}
				rx_left--;
				recv++;
			}

			if (!rx_left) {
				/* Next transfer */
				spin_lock_irqsave(&info->lock, flags);
				rx = ep93xx_spi_next(rx, msg);
				spin_unlock_irqrestore(&info->lock, flags);

				if (rx) {
					rx_left = rx->len;
					rx_buf = rx->rx_buf;
					if (rx->speed_hz) {
						/* FIXME: Change speed? */
						dev_dbg(info->dev,
							"transfer->speed_hz = %u\n",
							rx->speed_hz);
					}
				}
			}
		}
	}

	dev_dbg(info->dev, "sent = %d, received = %d\n", sent, recv);

	/* Wait for the Transmit FIFO to drain */
	while (!ep93xx_spi_tx_empty(info))
		;

	/* REVISIT: Do I need to empty the receive FIFO? */

	/* Disable the frame latch */
	if (info->chip->cs_control)
		info->chip->cs_control(0, info->chip->cs_active_high);

	msg->actual_length = sent;
	msg->complete(msg->context);
}

/*
 * Work queue handler to process all queued spi messages from
 * ep93xx_spi_transfer()
 */
static void ep93xx_spi_work(struct work_struct *work)
{
	struct ep93xx_spi *info = container_of(work, struct ep93xx_spi, work);
	unsigned long flags;

	spin_lock_irqsave(&info->lock, flags);
	while (!list_empty(&info->msg_queue)) {
		struct spi_message *msg;

		msg = container_of(info->msg_queue.next,
				struct spi_message, queue);
		list_del(&msg->queue);
		spin_unlock_irqrestore(&info->lock, flags);

		info->chip = spi_get_ctldata(msg->spi);

		ep93xx_spi_handle_message(info, msg);

		spin_lock_irqsave(&info->lock, flags);
	}
	spin_unlock_irqrestore(&info->lock, flags);
}

/*
 * spi_master->transfer() - adds a message to the controller's transfer queue.
 */
static int ep93xx_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
	struct ep93xx_spi *info = spi_master_get_devdata(spi->master);
	unsigned long flags;

	spin_lock_irqsave(&info->lock, flags);

	msg->actual_length = 0;

	list_add_tail(&msg->queue, &info->msg_queue);
	queue_work(info->workq, &info->work);

	spin_unlock_irqrestore(&info->lock, flags);

	return 0;
}

/*
 * Do initial allocation and setup for a slave device on the SPI bus.
 *
 * struct spi_board_info in the platform init code contains an array
 * of all the slave devices on the SPI bus, see include/linux/spi/spi.h
 * for more information.
 *
 * This driver uses the controller_data information to pass the ep93xx_spi_chip
 * information needed to access the slave device, for more information see
 * arch/arm/mach-ep93xx/include/mach/spi.h.
 */
static struct ep93xx_chip_data * ep93xx_alloc_chip_data(struct spi_device *spi)
{
	struct ep93xx_chip_data *chip;
	struct ep93xx_spi_chip *chip_info;
	int ret;

	dev_info(&spi->dev, "initial allocation for %s\n", spi->modalias);

	chip = kzalloc(sizeof(struct ep93xx_chip_data), GFP_KERNEL);
	if (!chip) {
		dev_err(&spi->dev, "failed: cannot allocate chip data");
		ret = -ENOMEM;
		goto error;
	}

	chip_info = spi->controller_data;
	if (chip_info) {
		if (chip_info->setup) {
			ret = chip_info->setup();
			if (ret) {
				dev_err(&spi->dev, 
					"failed: cannot setup hardware\n");
				goto error_free;
			}
		}
		
		/*
		 * NOTE: If a cs_control callback is not supplied the SSP
		 * controller will still use the SFRMOUT pin as the chip
		 * select. This pin is active LOW, and is asserted during 
		 * the entire transmission of the frame.
		 */
		if (chip_info->cs_control)
			chip->cs_control = chip_info->cs_control;
		
		if (chip_info->cleanup)
			chip->cleanup = chip_info->cleanup;
	}
		
	spi_set_ctldata(spi, chip);

	return chip;

error_free:
	kfree(chip);
error:
	return ERR_PTR(ret);
}

/* the spi->mode bits understood by this driver */
#define MODEBITS	(SPI_CPHA | SPI_CPOL | SPI_CS_HIGH)

/*
 * spi_master->setup() - updates the device mode and clocking records used
 *                       by a device's SPI controller; protocol code may
 *                       call this.
 *
 * This must fail if an unrecognized or unsupported mode is requested.
 * It's always safe to call this unless transfers are pending on the
 * device whose settings are being modified.
 */
static int ep93xx_spi_setup(struct spi_device *spi)
{
	struct ep93xx_spi *info = spi_master_get_devdata(spi->master);
	struct ep93xx_chip_data *chip;

	if (!spi->bits_per_word)
		spi->bits_per_word = 8;
	if (spi->bits_per_word < 4 || spi->bits_per_word > 16) {
		dev_err(&spi->dev,
			"failed setup: unsupported bits/word %d\n",
			spi->bits_per_word);
		return -EINVAL;
	}
	
	if (spi->mode & ~MODEBITS) {
		dev_err(&spi->dev,
			"failed setup: unsupported mode bits %x\n",
			spi->mode & ~MODEBITS);
		return -EINVAL;
	}

	/*
	 * Only alloc chip data on first setup for each slave device.
	 */
	chip = spi_get_ctldata(spi);
	if (!chip) {
		chip = ep93xx_alloc_chip_data(spi);
		if (IS_ERR(chip))
			return PTR_ERR(chip);
	}

	chip->max_speed_hz = spi->max_speed_hz;
	chip->bits_per_word = spi->bits_per_word;
	chip->cs_active_high = !!(spi->mode & SPI_CS_HIGH);
	chip->mode = spi->mode;
	
	ep93xx_apply_spi_setup(info, chip, spi->max_speed_hz);

	return 0;
}

/*
 * spi_master->cleanup() - frees controller-specific state
 */
static void ep93xx_spi_cleanup(struct spi_device *spi)
{
	struct ep93xx_chip_data *chip= spi_get_ctldata(spi);

	if (!chip)
		return;

	if (chip->cleanup)
		chip->cleanup();

	kfree(chip);
}

static int ep93xx_spi_probe(struct platform_device *pdev)
{
	struct ep93xx_spi *info = NULL;
	struct spi_master *master;
	struct resource *res;
	int err = 0;

	master = spi_alloc_master(&pdev->dev, sizeof(struct ep93xx_spi));
	if (!master) {
		dev_err(&pdev->dev, "failed to allocate spi_master\n");
		return -ENOMEM;
	}

	info = spi_master_get_devdata(master);
	info->master = spi_master_get(master);
	info->pdata = pdev->dev.platform_data;
	if (!info->pdata) {
		dev_err(&pdev->dev, "no platform data defined\n");
		err = -EINVAL;
		goto failed_free;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "failed to get I/O memory\n");
		err = -ENXIO;
		goto failed_free;
	}

	res = request_mem_region(res->start, resource_size(res), pdev->name);
	if (res == NULL) {
		dev_err(&pdev->dev, "failed to request I/O memory\n");
		err = -EBUSY;
		goto failed_free;
	}

	info->mmio_base = ioremap(res->start, resource_size(res));
	if (info->mmio_base == NULL) {
		dev_err(&pdev->dev, "failed to remap I/O memory\n");
		err = -ENXIO;
		goto failed_release_mem;
	}

	info->clk = clk_get(&pdev->dev, NULL);
	if (IS_ERR(info->clk)) {
		dev_err(&pdev->dev, "failed to get ssp output clock\n");
		err = PTR_ERR(info->clk);
		goto failed_iounmap;
	}

	info->dev = &pdev->dev;

	spin_lock_init(&info->lock);

	INIT_LIST_HEAD(&info->msg_queue);
	INIT_WORK(&info->work, ep93xx_spi_work);

	init_waitqueue_head(&info->waitq);

	info->workq = create_singlethread_workqueue(master->dev.parent->bus_id);
	if (!info->workq) {
		dev_err(&pdev->dev, "Unable to create work queue\n");
		goto failed_put_clk;
	}

	master->bus_num		= pdev->id;
	master->num_chipselect	= info->pdata->num_chipselects;
	master->setup		= ep93xx_spi_setup;
	master->transfer	= ep93xx_spi_transfer;
	master->cleanup		= ep93xx_spi_cleanup;

	clk_enable(info->clk);

	platform_set_drvdata(pdev, info);

	err = spi_register_master(master);
	if (err)
		goto failed_disable_clks;

	dev_info(&pdev->dev, "registered in bit-bang mode\n");

	return 0;

failed_disable_clks:
	platform_set_drvdata(pdev, NULL);
	clk_disable(info->clk);
failed_put_clk:
	clk_put(info->clk);
failed_iounmap:
	iounmap(info->mmio_base);
failed_release_mem:
	release_mem_region(res->start, resource_size(res));
failed_free:
	spi_master_put(info->master);
	return err;
}

static int ep93xx_spi_remove(struct platform_device *pdev)
{
	struct ep93xx_spi *info = platform_get_drvdata(pdev);
	struct resource *res= platform_get_resource(pdev, IORESOURCE_MEM, 0);

	ep93xx_spi_disable(info);
	spi_unregister_master(info->master);
	platform_set_drvdata(pdev, NULL);
	clk_disable(info->clk);
	clk_put(info->clk);
	iounmap(info->mmio_base);
	release_mem_region(res->start, resource_size(res));
	spi_master_put(info->master);

	return 0;
}

static struct platform_driver ep93xx_spidrv = {
	.driver		= {
		.name	= "ep93xx-spi",
		.owner	= THIS_MODULE,
	},
	.remove		= ep93xx_spi_remove,
};

static int __init ep93xx_spi_init(void)
{
        return platform_driver_probe(&ep93xx_spidrv, ep93xx_spi_probe);
}

static void __exit ep93xx_spi_exit(void)
{
        platform_driver_unregister(&ep93xx_spidrv);
}

module_init(ep93xx_spi_init);
module_exit(ep93xx_spi_exit);

MODULE_DESCRIPTION("EP93XX SPI Driver");
MODULE_ALIAS("platform:ep93xx-spi");
MODULE_AUTHOR("Ryan Mallon <ryan@bluewatersys.com>, "
		"H Hartley Sweeten <hsweeten@visionengravers.com");
MODULE_LICENSE("GPL");
