/*
 * snapper_cl15.c
 * SoC audio on Snapper CL15 module
 *
 * Copyright (C) 2008 Bluewater Systems Ltd
 * Author: Ryan Mallon <ryan@bluewatersys.com>
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>

#include <asm/io.h>
#include <asm/hardware/snapper.h>
#include <mach/ep93xx-regs.h>

#include "../codecs/tlv320aic23.h"
#include "ep93xx-pcm.h"
#include "ep93xx-i2s.h"

/*
 * Possible clock rates are: 11110400, 11427840, 12000000, 12800000
 */
#define SNAPPER_CL15_AUDIO_CLOCK	11110400

static int snapper_cl15_startup(struct snd_pcm_substream *substream)
{
	return 0;
}

static void snapper_cl15_shutdown(struct snd_pcm_substream *substream)
{
}

static int snapper_cl15_hw_params(struct snd_pcm_substream *substream,
				  struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->dai->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->dai->cpu_dai;
	int ret;
	
	ret = snd_soc_dai_set_fmt(codec_dai, SND_SOC_DAIFMT_I2S |
				  SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_fmt(cpu_dai, SND_SOC_DAIFMT_I2S |
				  SND_SOC_DAIFMT_CBS_CFS);
	if (ret < 0)
		return ret;

	ret = snd_soc_dai_set_sysclk(codec_dai, 0, SNAPPER_CL15_AUDIO_CLOCK, 
				     SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;
	
	ret = snd_soc_dai_set_sysclk(cpu_dai, 0, SNAPPER_CL15_AUDIO_CLOCK, 
				     SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;
	return 0;
}

static struct snd_soc_ops snapper_cl15_ops = {
	.startup	= snapper_cl15_startup,
	.shutdown	= snapper_cl15_shutdown,
	.hw_params	= snapper_cl15_hw_params,
};

static const struct snd_soc_dapm_widget tlv320aic23_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone Jack", NULL),
	SND_SOC_DAPM_LINE("Line In", NULL),
	SND_SOC_DAPM_MIC("Mic Jack", NULL),
};

static const struct snd_soc_dapm_route audio_map[] = {
	{"Headphone Jack",	NULL,	"LHPOUT"},
	{"Headphone Jack",	NULL,	"RHPOUT"},
	
	{"LLINEIN",		NULL,	"Line In"},
	{"RLINEIN",		NULL,	"Line In"},

	{"MICIN",		NULL,	"Mic Jack"},
};

static int snapper_cl15_tlv320aic23_init(struct snd_soc_codec *codec)
{
	
	printk(KERN_ERR "%s - here\n", __FUNCTION__);
	snd_soc_dapm_new_controls(codec, tlv320aic23_dapm_widgets,
				  ARRAY_SIZE(tlv320aic23_dapm_widgets));
	
	snd_soc_dapm_add_routes(codec, audio_map, ARRAY_SIZE(audio_map));
	snd_soc_dapm_enable_pin(codec, "Headphone Jack");
	snd_soc_dapm_enable_pin(codec, "Line In");
	snd_soc_dapm_enable_pin(codec, "Mic Jack");

	snd_soc_dapm_sync(codec);
	return 0;
}

static struct snd_soc_dai_link snapper_cl15_dai = {
	.name		= "TLV320AIC23",
	.stream_name	= "AIC23",
	.cpu_dai	= &ep93xx_i2s_dai,
	.codec_dai	= &tlv320aic23_dai,
	.init		= snapper_cl15_tlv320aic23_init,
	.ops		= &snapper_cl15_ops,
};

static struct snd_soc_card snd_soc_card_snapper_cl15 = {
	.name		= "Snapper CL15",
	.platform	= &ep93xx_soc_platform,
	.dai_link	= &snapper_cl15_dai,
	.num_links	= 1,
};

static struct snd_soc_device snapper_cl15_snd_devdata = {
	.card		= &snd_soc_card_snapper_cl15,
	.codec_dev	= &soc_codec_dev_tlv320aic23,	
};

static struct platform_device *snapper_cl15_snd_device;

static int __init snapper_cl15_init(void)
{
	int ret;

	snapper_cl15_snd_device = platform_device_alloc("soc-audio", -1);
	if (!snapper_cl15_snd_device)
		return -ENOMEM;

	platform_set_drvdata(snapper_cl15_snd_device, 
			     &snapper_cl15_snd_devdata);
	snapper_cl15_snd_devdata.dev = &snapper_cl15_snd_device->dev;
	ret = platform_device_add(snapper_cl15_snd_device);
	if (ret)
		platform_device_put(snapper_cl15_snd_device);

	/* FIXME - Move to core? */
	{
		u32 val;
		preempt_disable();	
		val = __raw_readl(EP93XX_SYSCON_DEVCFG);
		val &= ~EP93XX_SYSCON_DEVCFG_I2SONSSP;
		val |= EP93XX_SYSCON_DEVCFG_I2SONAC97;
		val |= EP93XX_SYSCON_DEVCFG_A1ONG | EP93XX_SYSCON_DEVCFG_A2ONG;
		__raw_writel(0xaa, EP93XX_SYSCON_SWLOCK);
		__raw_writel(val, EP93XX_SYSCON_DEVCFG);
		preempt_enable();
	}

	return ret;
}

static void __exit snapper_cl15_exit(void)
{
	platform_device_unregister(snapper_cl15_snd_device);
}

module_init(snapper_cl15_init);
module_exit(snapper_cl15_exit);

MODULE_AUTHOR("Ryan Mallon <ryan@bluewatersys.com>");
MODULE_DESCRIPTION("ALSA SoC for Snapper CL15");
MODULE_LICENSE("GPL");

